#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-10-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000  #temp, finally 3000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_hue_chroma (image, drawable, uhue, uchroma, ulightness):    # Hue-Chroma  
    hue=uhue                #0.0   #value_range  (-180.0, 180.0) 
    chroma=uchroma          #12.0  #value_range  (-100.0, 100.0)  
    lightness=ulightness    #0.0   #value_range  (-100.0, 100.0)
    mystring="gegl:hue-chroma hue=%d chroma=%d lightness=%d"
    gegl_graph_string=mystring % (hue, chroma, lightness)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_saturation (image, drawable, uscale, ucolorspace):    # saturation 
    scale=uscale            #1.5  #user_range   (0.0, 2.0)
    colorspace=ucolorspace  #0    #value_range  (0, 1, 2)  Native, CIE-lab/lch, CIE-iuv
    mystring="gegl:saturation scale=%f colorspace=%s"
    gegl_graph_string=mystring % (scale, colorspace)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_exposure (image, drawable, ublack, uexpo):      
    black_level=ublack    # value_range (-0.1, 0.1)
    exposure=uexpo        # ui_range    (-10.0, 10.0)
    gegl_graph_string="gegl:exposure black-level=%f exposure=%f" % (black_level, exposure)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_unsharp_mask (image, drawable, usdev , uscale, uthreshold):      
    std_dev=usdev           #user_range   (0.0, 40.0)
    scale=uscale            #user_range   (0.0, 10.0)
    threshold=uthreshold    #user_range   (0.0, 1.0) 
    mystring="gegl:unsharp-mask std-dev=%f scale=%f threshold=%f"
    gegl_graph_string=mystring % (std_dev, scale, threshold)
    gegl_graph(image, drawable,gegl_graph_string)

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.05, 0.95, False) #instead of 0.10/0.90
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # take care of the existing mean in determining the theoretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = (128+meanl)/2   
    elif meanl > 170: thmean = (127+meanl)/2   
    else: thmean = 127.5    # regular
    # mod to get lighter goal
    minmean = thmean*1.05  
    maxmean = thmean*1.15  
    
    modified=False
    if meanl < minmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)


#---------------------------------------------------------------------------
# adjust values and saturation
#---------------------------------------------------------------------------
def adjustmeanSV (inImage, inLayer):
    # decompose HSV
    (HSVI,HL,SL,VL) = pdb.plug_in_decompose (inImage, inLayer, "HSV", True)
    #pdb.gimp_display_new (HSVI)
    HL=HSVI.layers[0]; SL=HSVI.layers[1]; VL=HSVI.layers[2];
    adjustmeanv (HSVI, SL)
    adjustmeanv (HSVI, VL)
    pdb.plug_in_recompose (HSVI, HL)
    return(inLayer)

#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectS2 (inImage, inLayer, widthheight, texture, 
                 blurval, fantacolours, gradient, adjsat, adjchroma,
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    OrigWH = inLayer.width+inLayer.height
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    if texture==True:
        # Apply Camouflage  #fx_camouflage 2,12,100,85,85,85,115,115,115,145,145,145,175,175,175
        pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2,
        "-v - -fx_camouflage 2,12,100,85,85,85,115,115,115,145,145,145,175,175,175 ") 
        camouflage = newImage.layers[0]
        camouflage.mode = LAYER_MODE_OVERLAY
        camouflage.opacity = 25.0    
        camouflage.name = "[GMIC] -fx_camouflage 2,12,100,(85*3),(115*3),(145*3),(175*3)"
        
        Preprocessed = pdb.gimp_image_merge_down (newImage, camouflage, CLIP_TO_IMAGE)
    # prepare image for effects Blur + USM 
    # - Blur  
    BlurSharp = Preprocessed.copy()
    newImage.add_layer (BlurSharp, 0)
    uRadius=float(blurval)
    BlurSharp.name = "Median Blurred "+str(uRadius) 
    gegl_median_blur (newImage, BlurSharp, uRadius)
    # - USM
    uSdev=float(blurval); uScale=1.0; uThreshold=0.0       # new parameters
    BlurSharp.name = "Median Blur "+str(uSdev)+" then USM "+str(uSdev)
    gegl_unsharp_mask (newImage, BlurSharp, uSdev, uScale, uThreshold)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Effect: Diff Tensors  fx_diffusiontensors 8,16,2,1,0,1,0,3,0,50,50
    pdb.plug_in_gmic_qt(newImage, BlurSharp, 1,2 ,
    "-v - -fx_diffusiontensors 8,16,2,1,0,1,0,3,0,50,50 ")    
    Diff_Tensors2 = newImage.layers[0]
    Diff_Tensors2.mode = LAYER_MODE_OVERLAY # chg v2 from SOFTLIGHT 
    Diff_Tensors2.opacity = fantacolours*1.0  
    Diff_Tensors2.name = "[GMIC] fx_diffusiontensors 8,16,2,1,0,1,0,3,0,50,50 "
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Effect Bokeh (v2 added)
    Bokeh = pdb.gimp_layer_new_from_visible(newImage, newImage, "Bokeh: Layer for Bokeh effect")
    Bokeh.opacity = 100.00
    Bokeh.mode = LAYER_MODE_NORMAL  
    newImage.add_layer(Bokeh, 0)
    Bokeh = newImage.active_layer
    # WARNING! the Bokeh filter works ONLY if the underlying layer is in LAYER_MODE_NORMAL
    #          and 100.0% OPACITY !!!
    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    pdb.plug_in_gmic_qt(newImage, Bokeh, 1,0 ,
    #[G'MIC] Bokeh: fx_bokeh 3,8,0,256,8,0,0,0.2,210,210,80,160,1,30,20,20,1,2,170,130,20,110,0.15,0,50,50     
    "-v - -fx_bokeh 3,8,0,256,8,0,0,0.2,210,210,80,160,1,30,20,20,1,2,170,130,20,110,0.15,0,50,50 ")
    Bokeh = newImage.active_layer
    # NOW we set opacity and mode as we like
    Bokeh.opacity = min(100.0, fantacolours*2.0)      
    Bokeh.mode = LAYER_MODE_OVERLAY      
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # combine
    Diff_Tensors_Bokeh = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Diff_Tensors_Bokeh, 0)
    Diff_Tensors_Bokeh.name = "Diff_Tensors_Bokeh "
    Diff_Tensors_Bokeh.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Effect: Graphic Novel
    pdb.plug_in_gmic_qt(newImage, BlurSharp, 1,2 ,
    #[G'MIC] Graphic Novel: fx_graphic_novelfxl 0,2,6,5,40,0,4,200,0,1,2,0,10,0,0,0,1,1,1,2,0,10
    "-v - -fx_graphic_novelfxl 0,2,6,5,40,0,4,200,0,1,2,0,10,0,0,0,1,1,1,2,0,10 ")
    Graphic_Novel = newImage.layers[0]
    Graphic_Novel.mode = LAYER_MODE_PIN_LIGHT  
    Graphic_Novel.opacity = 35.0     
    GNparms = "0,2,6,5,40,0,4,200,0,1,2,0,10,0,0,0,1,1,1,2,0,10"
    Graphic_Novel.name = "[GMIC] -fx_graphic_novelfxl "+GNparms
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # prepare first base  
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base for Aurora "
    Base1.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # Apply Aurora   gcd_aurora 6,1,0
    pdb.plug_in_gmic_qt(newImage, Base1, 1,2 ,
    #[G'MIC] Aurora: gcd_aurora 6,1,0
    "-v - -gcd_aurora 6,1,0 ")
    Aurora = newImage.layers[0]
    Aurora.mode = LAYER_MODE_OVERLAY    
    Aurora.opacity = 50.0  
    Aurora.name = "[GMIC] -gcd_aurora 6,1,0"
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # add a gradient
    Gradient = pdb.gimp_layer_new (newImage, width, height, RGB_IMAGE,   
                                  "Gradient Layer "+gradient, 20.0, LAYER_MODE_OVERLAY)   
    newImage.add_layer(Gradient, 0)
    pdb.gimp_context_set_background((255,255,255,255))
    pdb.gimp_edit_fill (Gradient, FILL_BACKGROUND)
    pdb.gimp_context_set_gradient (gradient)
     
    pdb.gimp_edit_blend (Gradient, CUSTOM_MODE, LAYER_MODE_NORMAL,
                         GRADIENT_SHAPEBURST_DIMPLED,
                         100.0, 0, REPEAT_NONE, True, False, 1, 1, False,  
                         width/2, height/2, width, height)   
    
    pdb.plug_in_gauss_iir2 (newImage, Gradient, 500, 500)   
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare second base  
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Base for Canvas "
    Base2.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # combine with original
    OriginalOnTop = Original.copy()
    newImage.add_layer (OriginalOnTop, 0)
    pdb.gimp_item_set_visible (OriginalOnTop, True)
    OriginalOnTop.mode = LAYER_MODE_NORMAL
    OriginalOnTop.opacity = 50.0     
    OriginalOnTop.name = "Original on Top "
    mask = pdb.gimp_layer_create_mask (OriginalOnTop, ADD_MASK_WHITE)    
    pdb.gimp_layer_add_mask (OriginalOnTop, mask)
    pdb.gimp_drawable_curves_spline (mask, HISTOGRAM_VALUE, 4, (0.0,0.00,1.0,0.5))  #white to gray mask
    
    #pdb.gimp_display_new (newImage) 
    #return
        
    # prepare third base  
    Base3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base3, 0)
    Base3.name = "Base for Colours and Canvas "
    Base3.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Apply colours adjustments (user choice)
    if adjsat!=1.00:      
        SCALE=float(adjsat); COLORSPACE="native"
        gegl_saturation (newImage, Base3, SCALE, COLORSPACE)
    if adjchroma!=0.00:        
        UHUE=0.0; UCHROMA=float(adjchroma); ULIGHTNESS=0.0
        gegl_hue_chroma (newImage, Base3, UHUE, UCHROMA, ULIGHTNESS)
        
    # Apply Final Canvas  
    pdb.plug_in_gmic_qt(newImage, Base3, 1,0,
    "-v - -texturize_canvas 20,4,0.8 ")
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Pre-outcome to check luminosity
    Pre_outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
    newImage.add_layer (Pre_outcome, 0)
    Pre_outcome.name = "Preliminary outcome to check luminosity"
    # adjust mean values
    adjustmeanv (newImage, Pre_outcome)  
    Pre_outcome.mode = LAYER_MODE_MULTIPLY
    Pre_outcome.opacity = 35.0
    
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome"
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect S2 (Fantasy Graphic Effect), parms= "
    outcome.name = outcome.name+str(int(texture))
    outcome.name = outcome.name+", "+str(int(blurval))
    outcome.name = outcome.name+", "+str(int(fantacolours))
    outcome.name = outcome.name+", "+gradient
    outcome.name = outcome.name+", "+str(adjsat)
    outcome.name = outcome.name+", "+str(adjchroma)
    
    if widthheight==0: widthheight=OrigWH
    uRatio = widthheight*1.00/FixWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    #info_msg ("SCALING")   # not during tests, scaling is slow
    pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect S2 (Fantasy Graphic Effect - usually not for people) "

register(
    "JvidEffectS2",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_TOGGLE, "texture", "Texturize Source Image?", True),
        (PF_ADJUSTMENT, "blurval", "Blur value", 3, (1, 15, 1)),
        (PF_ADJUSTMENT, "fantacol", "Plus Fantasy",
                                    50, (1, 99, 1)),
        (PF_GRADIENT, "gradient", "Gradient name", "Pastel Rainbow"),
        (PF_ADJUSTMENT, "adjsat", "Adjust final Saturation "+
                                  "\n (0.65-0.99 to reduce it, 1.01-1.50 to increase it)",
                                   1.00, (0.65, 1.50, 0.01)),
        (PF_ADJUSTMENT, "adjchroma", "Adjust final Chroma "+
                                  "\n (neg value to reduce it, pos value to increase it)",
                                   0.00, (-20.0, 25.00, 0.50)),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectS2,
    menu=menu
)

main()
